* QL WORLD DIY TOOLKIT - ARRAY SEARCH ROUTINES
* Version 0.7, Copyright 1992 Simon N Goodwin.
*
initialise lea.l    define,a1
           movea.w  $110,a2         BP.INIT vector
           jmp      (a2)

define     dc.w     0,0             No procedures
           dc.w     2               One long-named function
           dc.w     inarray-*
           dc.b     8,'INARRAY%'
           dc.w     0

*
* INARRAY% ( array [ ,start% ] ,value ) scans string arrays FAST
*

inarray    cmpa.l   a3,a5           Any parameters?
           beq.s    bad_param
           cmpi.b   #3,0(a3,a6.l)   Initial array parameter?
           bne.s    bad_param
           moveq    #3,d5           Mask for known SB types
           and.b    1(a3,a6.l),d5   Pick up datatype
           beq.s    bad_param       Exclude substring type
*
* D5 = type; A3 = Name Table offset of the array to be searched
*
           movea.l  4(a3,a6.l),a1   A1 -> Array descriptor
           movea.l  40(a6),a4       Fetch BV.VVBAS offset
           adda.l   a4,a1           A1 is descriptor A6 offset
           adda.l   0(a1,a6.l),a4   A4 is offset of values
           move.w   4(a1,a6.l),d0   Get number of dimensions
           moveq    #1,d7           Initial index multiplier
           cmpi.b   #1,d5           Check type; 1 means String
           bne.s    bad_param
*
* Treat two-dimensional character arrays as a special case
*
           subq.w   #2,d0           Faster than CMP if volatile
           bne.s    bad_param       Wrong number of dimensions
           move.w   6(a1,a6.l),d7   Maximum element number
           move.w   8(a1,a6.l),d5   Stride of each string, 4+ 
           bpl.s    next_param      Disallow negative lengths
           bra.s    bad_param 

next_param addq.l   #8,a3
           cmpa.l   a3,a5           Any more parameters?
           beq.s    bad_param
           swap     d7              Save count for later
           move.w   d5,d7           Save type: 2=F, 3=%, 4+=$
*
* D7.W = type; D7.H = last element, 0-32767; A4 = initial offset
*
           moveq    #0,d5           Assumed start, element zero
           lea.l    8(a3),a2        Delimit one Name Table entry
           cmpa.l   a2,a5           Is there one parameter left?
           beq.s    strings         If so, D5 is set correctly
*
* Fetch the optional integer 'first element to scan' in D5.W
*
           move.l   a5,d5           Save end of parameters
           movea.l  a2,a5           Delimit the first parameter
           movea.w  $112,a2         Read vector to get integers
           jsr      (a2)            Call CA.GTINT
           bne.s    bad_exit 
           move.w   0(a1,a6.l),d0   Pick up the value
           bpl.s    good_param      It needs to be positive
*
bad_param  moveq    #-15,d0         Signal a BAD PARAMETER error
bad_exit   rts                      Return the error code in D0
*
good_param addq.l   #2,$58(a6)      Unstack integer from BV.RIP
           movea.l  a5,a3           Step past this parameter
           movea.l  d5,a5           Restore end of list
           move.w   d0,d5           Update D5 to match parameter
           lea.l    8(a3),a2        Only one parameter left?
           cmpa.l   a2,a5
           bne.s    bad_param

*
* Search string array elements for the last parameter
*
strings    movea.w  $116,a2         Vector to get strings
           jsr      (a2)            CA.GTSTR
           bne.s    exit
           moveq    #1,d0           Allow for the odd byte
           add.w    0(a1,a6.l),d0   Unstack the text length
           bclr     #0,d0           D0 is length rounded up
           add.l    d0,$58(a6)      Leave one word for result
           moveq    #0,d6
           move.w   d7,d6           D6 is string stride
           move.w   d5,d4           Work out start offset
           mulu     d6,d4           Scale for elements
           adda.l   d4,a4           Advance to first element
           movea.l  a4,a0           Set A0 for UT.CSTR
           swap     d7              Now D7 is element count
search_str cmp.w    d5,d7           Are we past the end?
           bcs.s    finished
           movem.l  d5/d7/a1/a0,-(a7)
           moveq    #-7,d0
           moveq    #0,d5
           move.w   0(a1,a6.l),d5
           cmp.w    0(a0,a6.l),d5
           bne.s    no_match
           subq.w   #1,d5
test_str   move.b   2(a1,a6.l),d7
           cmp.b    2(a0,a6.l),d7
           bne.s    no_match
           addq.l   #1,a1
           addq.l   #1,a0
           dbf      d5,test_str
           moveq    #0,d0           A match has been found!!
no_match   movem.l  (a7)+,d5/d7/a1/a0
           tst      d0
           beq.s    stack_int       Return D5 if they match
           addq.w   #1,d5
           adda.l   d6,a0           Next; Note implicit EXT.L
           bra.s    search_str
*
* Return location 0-32767, or -7 if the value was not found
*
finished   moveq    #-7,d5          NOT FOUND Qdos error code
stack_int  move.l   $58(a6),a1      Retrieve BV.RIP
           move.w   d5,0(a1,a6.l)   Stack a one word reply
           moveq    #3,d4           Indicate integer result
           moveq    #0,d0           Signal ERR.OK, it worked
exit       rts

           end
